/**************************************************************
***************************************************************
*                                                             *
* All relevant (system) settings are done in configurations.h *
*                                                             *
***************************************************************
**************************************************************/


/****************************************************************************
 Include Section
*****************************************************************************/

#include "configurations.h"
#if defined INSTR_CACHE_EN

#include "exception.h"

#include <cplb.h>
#include <def_LPBlackfin.h>
#include "ndef_LPBlackfin.h"

#include "macros.h"
#include "BfMemMap.h"

/*****************************************************************************
 Prototypes
******************************************************************************/

.GLOBAL __icplb_init;
.GLOBAL __icplb_handler;
//.GLOBAL _dcplb_handler.cplb_miss;
.GLOBAL __iflushinv;
.GLOBAL __itest;
.GLOBAL __icachelockbyway;

/*****************************************************************************
 Global Variables
******************************************************************************/

.GLOBAL _b4_icplb_data_pointer;
.GLOBAL _b4_icplb_address_pointer;

/*****************************************************************************
 Variables
******************************************************************************/

.SECTION cplb_data;

.ALIGN 4;
.BYTE4 _b4_icplb_data_pointer = 0;
.BYTE4 _b4_icplb_address_pointer = 0;

/*****************************************************************************
 Symbolic constants / definitions
******************************************************************************/

/* number of entries in _b4_icplbs_table is depending on the PAGE_SIZE */
#if defined LIMIT32kB
    #define     PAGE_SIZE_VAL           PAGE_SIZE_1KB
#else
    #define     PAGE_SIZE_VAL           PAGE_SIZE_4MB
#endif

#if   PAGE_SIZE_VAL == PAGE_SIZE_4MB
#define         PAGE_LENGTH             0x00400000
#elif PAGE_SIZE_VAL == PAGE_SIZE_1MB
#define         PAGE_LENGTH             0x00200000
#elif PAGE_SIZE_VAL == PAGE_SIZE_4KB
#define         PAGE_LENGTH             0x00001000
#elif PAGE_SIZE_VAL == PAGE_SIZE_1KB
#define         PAGE_LENGTH             0x00000400
#else
#error PAGE_SIZE undefined
#endif

/* missing definitions */
#define  CPLB_MEM_LEV 0x00000200
#define nCPLB_MEM_LEV 0x0


/***************************************************************************************
L1 Instruction Memory Control Register (IMEM_CONTROL)
RESERVED                [0]     1
ENICPLB                 [1]     Instruction CPLB enable
                   DISICPLB     0 - CLPBs disabled, minimal address checking only
                    ENICPLB     1 - CPLBs enabled
IMC                     [2]     Enable/Disable Instruction Cache functionality
                        nIMC    0 - Upper 16K byte of LI instruction memory configured as SRAM, also invalidates all cache lines if previously configured as cache
                         IMC    1 - Upper 16K byte of L1 instruction memory configured as cache
ILOC                [ 6: 3]     0 - Way not locked
                                1 - Way locked
                      nILOC     0000 - All ways not locked
                                1111 - All Ways locked ( ILOC0 | ILOC1 | ILOC2 | ILOC3 )
RESERVED            [12: 7]
LRUPRIORST             [13]     0 - LRU priority functionality is enabled
                                1 - All cached LRU priority bits (LRUPRIO) are cleared
RESERVED            [31:14]
***************************************************************************************/

#define IMEM_CONTROL_VAL        ( ENICPLB | IMC | nILOC | nLRUPRIORST ) /* IMEM_CONTROL = 0x7 */


/***************************************************************************************
ICPLB Data Registers (ICPLB_DATAx)
CPLB_VALID              [0]
               CPLB_INVALID     0 - Invalid (disabled) CPLB entry
                 CPLB_VALID     1 - Valid (enabled) CPLB entry
CPLB_LOCK               [1]     Can be used by software in CPLB replacement algorithms
                CPLB_UNLOCK     0 - Unlocked, CPLB entry can be replaced
                  CPLB_LOCK     1 - Locked, CPLB entry should not be replaced
CPLB_USER_RD            [2]
              nCPLB_USER_RD     0 - User mode read access generates protection violation exception
               CPLB_USER_RD     1 - User mode read access permitted
RESERVED            [ 7: 3]
CPLB_LRUPRIO            [8]     <-> LRUPRIORST in IMEM_CONTROL
              nCPLB_LRUPRIO     0 - Low importance
               CPLB_LRUPRIO     1 - High importance
RESERVED           [11: 10]
CPLB_MEM_LEV            [9]     Determines line fill and write buffers. For derivatives with internal L2 memory only.
                                The state of this bit has no affect if this page describes L1 memory.
              nCPLB_MEM_LEV     0 - High priority (usually best for on-chip L2 memory pages)
               CPLB_MEM_LEV     1 - Low priority (usually best for off-chip L2 memory pages)
CPLB_L1_CHBL           [12]     Clear this bit whenever L1 memory is configured as SRAM
              nCPLB_L1_CHBL     0 - Non-cacheable in L1
               CPLB_L1_CHBL     1 - Cacheable in L1
RESERVED            [15:14]
PAGE_SIZE           [17:16]
              PAGE_SIZE_1KB     00 - 1K byte page size
              PAGE_SIZE_4KB     01 - 4K byte page size
              PAGE_SIZE_1MB     10 - 1M byte page size
              PAGE_SIZE_4MB     11 - 4M byte page size
RESERVED            [31:18]
***************************************************************************************/

/***************************************************************************************
* Bit 9 of the ICPLB_DATAx is shown as Reserved bit however, when ICACHE is used this  *
* bit needs to be set to the same state as bit 12 (CPLB_L1_CHBL) of this register as a *
* workaround to anomaly: 05000258. Bit 9 = CPLB_MEM_LEV                                *
***************************************************************************************/

#undef  CPLB_L1_CHBL
#define CPLB_L1_CHBL    0x00001200

/********************************************************************************
* There are two default CPLB descriptors for data accesses to the scratchpad    *
* data memory and to the system and core MMR space. These default descriptors   *
* define the above space as non-cacheable, so that additional CPLBs do not need *
* to be set up for these regions of memory.                                     *
* If valid CPLBs are set up for this space, the default CPLBs are ignored.      *
********************************************************************************/

/* ICPLB_DATA = 0x0003 0003 */
/* 4MB starting at 0xFFA0 0000 covers til 0xFFE0 0000 */
/* This would replace default CPLB descriptor */
/* ICPLB_DATA = 0x0002 0003 */
/* 1MB starting at 0xFFA0 0000 covers til 0xFFB0 0000 */
#define ICPLB_DATA_L1_VAL                       ( CPLB_VALID | CPLB_LOCK   | PAGE_SIZE_1MB | nCPLB_USER_RD | nCPLB_L1_CHBL | nCPLB_LRUPRIO )

#if defined (__ADSPBF54x__) || defined (__ADSPBF561__)
    /* ICPLB_DATA = 0x0002 1005 */
    #define ICPLB_DATA_L2_SRAM_CACHE_VAL        ( CPLB_VALID | CPLB_UNLOCK | PAGE_SIZE_1MB |  CPLB_USER_RD |  CPLB_L1_CHBL | nCPLB_LRUPRIO )
    /* ICPLB_DATA = 0x0002 0005 */
    #define ICPLB_DATA_L2_SRAM_NOCACHE_VAL      ( CPLB_VALID | CPLB_UNLOCK | PAGE_SIZE_1MB |  CPLB_USER_RD | nCPLB_L1_CHBL | nCPLB_LRUPRIO )
#endif

#if defined (__ADSPBF51x__) || defined (__ADSPBF52x__)
    /* ICPLB_DATA = 0x0002 1205 */
    #define ICPLB_DATA_BOOTROM_CACHE_VAL        ( CPLB_VALID | CPLB_UNLOCK | PAGE_SIZE_1MB |  CPLB_USER_RD |  CPLB_L1_CHBL | nCPLB_LRUPRIO )
    #define ICPLB_DATA_BOOTROM_NOCACHE_VAL      ( CPLB_VALID | CPLB_UNLOCK | PAGE_SIZE_1MB |  CPLB_USER_RD | nCPLB_L1_CHBL | nCPLB_LRUPRIO )
#endif
#if defined __ADSPBF54x__
    /* ICPLB_DATA = 0x0001 1205 */
    #define ICPLB_DATA_BOOTROM_CACHE_VAL        ( CPLB_VALID | CPLB_UNLOCK | PAGE_SIZE_4KB |  CPLB_USER_RD |  CPLB_L1_CHBL | nCPLB_LRUPRIO )
    #define ICPLB_DATA_BOOTROM_NOCACHE_VAL      ( CPLB_VALID | CPLB_UNLOCK | PAGE_SIZE_4KB |  CPLB_USER_RD | nCPLB_L1_CHBL | nCPLB_LRUPRIO )
#endif

/* ADSP-BF54x can handle more than 4MB of Async Mem !!! */
/* ICPLB_DATA = 0x0003 1203 */
#define ICPLB_DATA_ASYNC_CACHE_VAL              ( CPLB_VALID | CPLB_UNLOCK | PAGE_SIZE_4MB |  CPLB_USER_RD |  CPLB_L1_CHBL | nCPLB_LRUPRIO )
/* ICPLB_DATA = 0x0003 0003 */
#define ICPLB_DATA_ASYNC_NOCACHE_VAL            ( CPLB_VALID | CPLB_UNLOCK | PAGE_SIZE_4MB |  CPLB_USER_RD | nCPLB_L1_CHBL | nCPLB_LRUPRIO )

/* ICPLB_DATA = 0x0003 1205 */
#define ICPLB_DATA_SDRAM_CACHE_VAL              ( CPLB_VALID | CPLB_UNLOCK | PAGE_SIZE_VAL |  CPLB_USER_RD |  CPLB_L1_CHBL | nCPLB_LRUPRIO )
/* ICPLB_DATA = 0x0003 0005 */
#define ICPLB_DATA_SDRAM_NOCACHE_VAL            ( CPLB_VALID | CPLB_UNLOCK | PAGE_SIZE_VAL |  CPLB_USER_RD | nCPLB_L1_CHBL | nCPLB_LRUPRIO )


#define ICPLB_DATA_SDRAM_NULLPTR_PROT_VAL       ( CPLB_VALID | CPLB_LOCK   | PAGE_SIZE_1KB |  CPLB_USER_RD | nCPLB_L1_CHBL | nCPLB_LRUPRIO )


/*****************************************************************************
 ICPLB Data
******************************************************************************/

.SECTION cplb_data;
.ALIGN 4;


_icplbs_table:

.BYTE4 _b4_icplbs_table[]=


    INSTR_SRAM_START,ICPLB_DATA_L1_VAL,


    #if defined (__ADSPBF54x__) || defined (__ADSPBF561__)
//        L2_SRAM_START,ICPLB_DATA_L2_SRAM_CACHE_VAL,
        L2_SRAM_START,ICPLB_DATA_L2_SRAM_NOCACHE_VAL,
    #endif


#if ( SDRAM_SIZE == 128 || SDRAM_SIZE == 64 || SDRAM_SIZE == 32 || SDRAM_SIZE == 16 ) || defined ( LIMIT32kB )

    #if defined NULLPTR_PROTECTION
        SDRAM_START+PAGE_LENGTH*0,ICPLB_DATA_SDRAM_NULLPTR_PROT_VAL,
        SDRAM_START+PAGE_LENGTH*0,ICPLB_DATA_SDRAM_CACHE_VAL|CPLB_LOCK,
    #else
        SDRAM_START+PAGE_LENGTH*0,ICPLB_DATA_SDRAM_CACHE_VAL,
    #endif
    SDRAM_START+PAGE_LENGTH*1,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*2,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*3,ICPLB_DATA_SDRAM_CACHE_VAL,
    /* --> 16MB */
    /* --> 4kB */

#endif


#if ( SDRAM_SIZE == 128 || SDRAM_SIZE == 64 || SDRAM_SIZE == 32 ) || defined ( LIMIT32kB )

    SDRAM_START+PAGE_LENGTH*4,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*5,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*6,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*7,ICPLB_DATA_SDRAM_CACHE_VAL,
    /* --> 32MB */
    /* --> 8kB */

#endif


#if ( SDRAM_SIZE == 128 || SDRAM_SIZE == 64 ) || defined ( LIMIT32kB )

    SDRAM_START+PAGE_LENGTH*8,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*9,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*10,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*11,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*12,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*13,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*14,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*15,ICPLB_DATA_SDRAM_CACHE_VAL,
    /* --> 64MB */
    /* --> 16kB */

#endif


#if ( SDRAM_SIZE == 128 ) || defined ( LIMIT32kB )

    SDRAM_START+PAGE_LENGTH*16,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*17,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*18,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*19,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*20,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*21,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*22,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*23,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*24,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*25,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*26,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*27,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*28,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*29,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*30,ICPLB_DATA_SDRAM_CACHE_VAL,
    SDRAM_START+PAGE_LENGTH*31,ICPLB_DATA_SDRAM_CACHE_VAL,
    /* --> 128MB */
    /* --> 32kB */

#endif


    #if defined (__ADSPBF51x__) || defined (__ADSPBF52x__) || defined (__ADSPBF54x__)
        BOOT_ROM_START,ICPLB_DATA_BOOTROM_CACHE_VAL,
//        BOOT_ROM_START,ICPLB_DATA_BOOTROM_NOCACHE_VAL,
    #endif


    ASYNC_MEM_START,ICPLB_DATA_ASYNC_CACHE_VAL,
//    ASYNC_MEM_START,ICPLB_DATA_ASYNC_NOCACHE_VAL,


    /* end of section - termination */
    -1;

_icplbs_table.end:


/*****************************************************************************
 ICPLB Code
******************************************************************************/

/*****************************************************************************
 Functions
******************************************************************************/

.SECTION cplb_code;

/*****************************************************************************
 Function:              __icplb_init
 Description:           initialize instruction CPLBs
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        ASTAT,R7:5,P5:4,DAG2:0
 Global Variables Used: _b4_icplb_address_pointer
                        _b4_icplb_data_pointer
 C-Callable :           no
*****************************************************************************/

__icplb_init:
    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:5,P5:4);
    PUSH_DAG_REGS(0);
    PUSH_DAG_REGS(1);
    PUSH_DAG_REGS(2);
    IMM32(P5,CORE_MMR_BASE);


    M0 = 0(z); M1 = 0(z); M2 = 0(z);
    L0 = 0(z); L1 = 0(z); L2 = 0(z);
    B0 = 0(z); B1 = 0(z); B2 = 0(z);


    /* RESET IMEM_CONTROL register */
    R7 = 0x1 (z);
    [P5 + lo(IMEM_CONTROL)] = R7;
    ssync;

    /* Initialize ICPLBs */
    IMM32(I0,ICPLB_ADDR0);
    IMM32(P4,_b4_icplb_address_pointer); /* save address pointer for CPBL exception handler required for Round Robin Scheduling */
    R6 = I0;
    [P4] = R6;

    IMM32(I1,ICPLB_DATA0);
    IMM32(P4,_b4_icplb_data_pointer); /* save data pointer for CPBL exception handler required for Round Robin Scheduling */
    R6 = I1;
    [P4] = R6;
    
    IMM32(I2,_b4_icplbs_table);
    R5 = -1; /* end point comparison */
    R6 = 15; /* max counter */

    /* read entries from table */
    read_iaddr:
        R7 = [I2++];
        CC = R7 == R5;
        if CC jump idone;
        [I0++] = R7;
        csync;
    read_idata:
        R7 = [I2++];
        [I1++] = R7;
        csync;
        R6 = R6 + R5;
        CC = R6 == R5;
        if !CC jump read_iaddr;
    idone:

    /* ENABLE ICPLBs */
    R7 = IMEM_CONTROL_VAL;
    [P5 + lo(IMEM_CONTROL)] = R7;
    ssync;


    icplb_init.done:
    POP_DAG_REGS(2);
    POP_DAG_REGS(1);
    POP_DAG_REGS(0);
    (R7:5,P5:4) = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;
__icplb_init.END:


/*****************************************************************************
 Function:              __icplb_handler
 Description:           genereric instruction CPLB handler
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        ASTAT,R7:4
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

__icplb_handler:
    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:4);

    cli R4;

    R7 = SEQSTAT;
    R6 = EX_BITS;
    R5 = R7 & R6;

    /**********************
    * Verify a ICPLB miss *
    **********************/
    R6 = EX_SYS_CCPLBMISS;
    CC = R5 == R6;
    if CC jump icplb_handler.cplb_miss (BP);

    EMUEXCPT;
    jump (0);

    icplb_handler.cplb_miss:
    call _icplb_handler.cplb_miss;
//    jump icplb_handler.done;


    icplb_handler.done:
    bitclr(R4,bitpos(EVT_IVG14));
    sti R4;
    (R7:4) = [SP++];
    ASTAT = [SP++];
    unlink;
    rti;
__icplb_handler.end:


/*****************************************************************************
 Function:              _icplb_handler.cplb_miss
 Description:           instruction access CPLB handler
                        Identify correct ICPLB entry in ICPLB table
                        Replace one (Round Robin Scheduling) with the correct entry
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        ASTAT,R7:2,P5:3,DAG1:0
 Global Variables Used: _b4_icplb_address_pointer
                        _b4_icplb_data_pointer
 C-Callable :           no
*****************************************************************************/

_icplb_handler.cplb_miss:
    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:1,P5:3);
    PUSH_DAG_REGS(0);
    PUSH_DAG_REGS(1);
    IMM32(P5,CORE_MMR_BASE);


    /*************************************
    * Check, that access was done for    *
    * an available/populated memory area *
    *************************************/

    R6 = [P5 + lo(ICPLB_FAULT_ADDR)];      /* ICPLB_FAULT_ADDR */
    IMM32(R7,SDRAM_LENGTH);
    CC = R6 < R7;
    if CC jump icplb_handler.cplb_miss.icplb_exchange (BP);
    IMM32(R7,ASYNC_MEM_START);
    CC = R6 < R7;
    if !CC jump icplb_handler.cplb_miss.icplb_exchange (BP);

    icplb_handler.cplb_miss.reserved_mem_access:
    EMUEXCPT;
    jump (0);

    icplb_handler.cplb_miss.icplb_exchange:
    /************************************************
    * Load starting point in _b4_icplbs_table for:  *
    * ICPLB_ADDR -> I0                              *
    * ICPLB_DATA -> I1                              *
    ************************************************/

    IMM32(I0,_b4_icplbs_table); /* ICPLB_ADDRx */
    B0 = I0;
    M0 = 4;
    L0 = 4*length(_b4_icplbs_table)-4;
    
    /*****************************************************
    * Search for the right entry in the _b4_icplbs_table *
    * depending on ICPLB_FAULT_ADDR                      *
    *****************************************************/

    R6 = [P5 + lo(ICPLB_FAULT_ADDR)]; /* ICPLB_FAULT_ADDR */
    IMM32(R1,0x00030000); /* for PAGE_SIZE bits */

    icplb_handler.find_ICPLB_entry:
    R5 = [I0++];    /* ICPLB_ADDRx */
    R4 = [I0++];    /* ICPLB_DATAx */
    /* Extract PAGE_SIZE bits */
    R3 = R4 & R1;
    R3 = R4 >> 16;  /* R3 = CPLB Page Size bits */


    /*****************
    * find PAGE_SIZE *
    *****************/

    CC = R3 == 0x3;
    if CC jump MB4 (BP);    /* PAGE_SIZE = 4MB */
    CC = R3 == 0x2;
    if CC jump MB1;         /* PAGE_SIZE = 1MB */
    CC = R3 == 0x1;
    if CC jump kB4;         /* PAGE_SIZE = 4kB */
                            /* else: PAGE_SIZE = 1kB */

    kB1:                    /* 1kB: 0x0400 */
    R2 = 0x400;
    jump icplb_handler.page_size_End;

    kB4:                    /* 4kB: 0x1000 */
    R2 = 0x1000;
    jump icplb_handler.page_size_End;

    MB1:                    /* 1MB: 0x100000 */
    IMM32(R2,0x00100000);
    jump icplb_handler.page_size_End;

    MB4:                    /* 4MB: 0x400000 */
    IMM32(R2,0x00400000);

    icplb_handler.page_size_End:
    R7 = R5 + R2;           /* Store ICPLB_ADDRx end address */


    /***********************************************************
    * Check if actual entry in _b4_icplbs_table is valid       *
    * ICPLB_ADDRx_START <= ICPLB_FAULT_ADDR < ICPLB_ADDRx_END  *
    * R5 <= R6 < R7                                            *
    ***********************************************************/

    CC = R5 <= R6;
    if !CC jump icplb_handler.find_ICPLB_entry (BP);
    CC = R6 < R7;
    if !CC jump icplb_handler.find_ICPLB_entry (BP);


    /*************************************************
    * Load last index point in _b4_icplbs_table for: *
    * _b4_icplb_address_pointer -> I2                *
    * _b4_icplb_data_pointer -> I3                   *
    * both need to be updated at the end             *
    * initialization in: __icplb_init                *
    *************************************************/

    IMM32(B0,ICPLB_ADDR0); /* 0xFFE0 0200 */
    IMM32(P4,_b4_icplb_address_pointer);
    R7 = [P4];
    I0 = R7;
    M0 = 4;                     /* 4 byte steps */
    L0 = 64;                    /* 16 ICPLB_ADDRx entrties with each 4 bytes */

    IMM32(B1,ICPLB_DATA0); /* 0xFFE0 0100 */
    IMM32(P3,_b4_icplb_data_pointer);
    R7 = [P3];
    I1 = R7;
    M1 = M0;                    /* 4 byte steps */
    L1 = L0;                    /* 16 ICPLB_DATAx entrties with each 4 bytes */


    /****************************************
    * Check if actual ICPLB entry is locked *
    * If yes than try next one              *
    ****************************************/

    icplb_handler.check_ICPLB_LOCK:
    R2 = [I0++];                /* ICPLB_ADDRx */
    NOP;
    R3 = [I1++];                /* ICPLB_DATAx */
    CC = bittst(R3,1);
    if CC jump icplb_handler.check_ICPLB_LOCK;
    

    /* decrement index because the pointer is already at the next entry */
    R2 = [I0--];                /* ICPLB_ADDRx */
    R3 = [I1--];                /* ICPLB_DATAx */


    /**********************
    * Replace ICPLB entry *
    **********************/

    /* Before loading new descriptor data into any CPLBs, the corresponding group
    of sixteen CPLBs must be disabled using the Enable ICPLB (ENICPLB) bit in the
    IMEM_CONTROL register for data descriptors */
    R7 = [P5 + lo(IMEM_CONTROL)];
    bitclr (R7,bitpos(ENICPLB));
    [P5 + lo(IMEM_CONTROL)] = R7;
    ssync;

    /* Update ICPLB entry and store address in _b4_dcplb_address_pointer and _b4_dcplb_data_pointer for Round Robin scheduling */
    [I0++] = R5;
    csync;
    R6 = I0;
    [P4] = R6;
    [I1++] = R4;
    csync;
    R6 = I1;
    [P3] = R6;

    /* ENABLE ICPLBs */
//    R7 = IMEM_CONTROL_VAL;
    bitset (R7,bitpos(ENICPLB));
    [P5 + lo(IMEM_CONTROL)] = R7;
    ssync;


    IMM32(P5,IMASK);
    R7 = [P5];
    bitclr (R7,bitpos(EVT_IVG14));
    [P5] = R7;


    icplb_handler.cplb_miss.done:
    POP_DAG_REGS(1);
    POP_DAG_REGS(0);
    (R7:1,P5:3) = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;
_icplb_handler.cplb_miss.end:


/*****************************************************************************
 Function:              __iflushinv
 Description:           Invalidate instruction cache
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        P5,R7
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

__iflushinv:
    link 0;
    [--SP] = (R7:7,P5:5);

    ssync;
    IMM32(P5,IMEM_CONTROL);
    R7 = [P5];
    bitclr(R7,bitpos(IMC));
    [P5] = R7;
    ssync;
    bitset(R7,bitpos(IMC));
    [P5] = R7;
    ssync;

    (R7:7,P5:5) = [SP++];
    unlink;
    rts;
__iflushinv.end:


/*****************************************************************************
 Function:              __itest
 Description:           Execute ITEST_COMMAND
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        P5:4,R7,LOOPREG
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

__itest:
    link 0;
    [--SP] = (R7:7,P5:5);
    IMM32(P5,CORE_MMR_BASE);



    (R7:7,P5:5) = [SP++];
    unlink;
    rts;
__itest.end:


/*****************************************************************************
 Function:              __icachelockbyway
 Description:           sets the ILOC bits in IMEM_CONTROL register
 Input Parameters:      e.g. R0 = ILOC1|ILOC2|ILOC3
                        R0 contains the bits for locking the icache ways
 Return Parameters:     none
 Registers Used:        P5,R7
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

__icachelockbyway:
    link 0;
    [--SP] = (R7:6,P5:5);

    IMM32(P5,IMEM_CONTROL);
    R7 = [P5];
    R6 = 1;
    [P5] = R6;
    ssync;
    bitclr(R7,bitpos(ILOC0));
    bitclr(R7,bitpos(ILOC1));
    bitclr(R7,bitpos(ILOC2));
    bitclr(R7,bitpos(ILOC3));
    R7 = R7|R0;
    [P5] = R7;
    ssync;

    (R7:6,P5:5) = [SP++];
    unlink;
    rts;
__icachelockbyway.end:


#endif /* defined INSTR_CACHE_EN */


/*****************************************************************************
 EOF
******************************************************************************/
